
#include "upf/upf.h"
#include "private.h"
#include "direnum.h"

namespace upf { namespace impl {

upf_CID alloc_new_cid()
{
    static upf_CID free_cid = 1;
    return free_cid++;
}


upf_CID classname_to_cid(const char *classname)
{
    return privateData->manager->classname2cid(classname);
}

upf_Object* create_instance(upf_CID classID)
{
    Ptr<IObject> obj = privateData->manager->createInstance(classID);
    if (obj)
    {
        upf_Object *abiObj = obj->_upf_getABIObject();
        abiObj->refCnt++; // will be decRef'ed by whoever called create_instance
        return abiObj;
    }
    else
        return NULL;
}


void Manager::_initClass()
{
    impl::ClassInfo *ci_impl = new impl::ClassInfo();
    // NB: must initialize the instance manually, because it wasn't created via
    //     a factory:
    ci_impl->_initInstance();
    Ptr<IWriteableClassInfo> ci = ci_impl;
    ClassRegistrator<impl::Manager>::prepare(ci, "upf::impl::Manager");
    Ptr<IClassInfo> frozen = ci->freeze();
    registerClass(frozen);

    _upf_initInstance();
}

Manager::Manager() 
{
}

Manager::~Manager()
{
    // NB: we must destroy the loaders in _reverse_ order of creation (i.e.
    //     destroy the first on added as the last one). The reason is that
    //     later loaders may use code that was loaded into program by previous
    //     ones. For example, any loader implemented in a DLL will depend on
    //     upf.impl.DllLoader and if we destroyed upf.impl.DllLoader instance
    //     first, it would unmap DLLs' code and any call to any function in
    //     any other loader would crash.
    //
    //     We have to release all ClassInfo and Factory objects from memory
    //     before unloading any object for the very same reason.
    
    m_classes.clear();
    
    while (!m_loaders.empty())
        m_loaders.pop_back();
}

bool Manager::registerClass(IClassInfo *classinfo)
{
    string name(classinfo->getName());
    CID cid(classinfo->getCID());
    if (m_classnames.find(name) != m_classnames.end())
        return false;
    m_classnames[name] = cid;

    ClassInfoData& dat = m_classes[cid];
    dat.classInfo = classinfo;
    dat.factory = classinfo->getFactory();
    // NB: there is tricky STL issue here: the following won't work:
    //            ClassInfoData dat;
    //            dat.classInfo = classinfo;
    //            dat.factory = classinfo->getFactory();
    //            m_classes[name] = dat;
    //     because m_classes[...] is *reference* and it would be assigned
    //     reference to 'dat', data would *not* be copied!
    
    if (classinfo->implementsInterface(InterfaceTraits<ILoader>::iid()))
    {
        Ptr<ILoader> loader = dat.factory->create();
        addLoader(loader);
    }

    return true;
}

CID Manager::classname2cid(const char *classname)
{
    ClassnamesTbl::iterator entry = m_classnames.find(classname);
    if (entry == m_classnames.end())
        return 0;
    else
        return entry->second;
}

Ptr<IClassInfo> Manager::getClassInfo(CID classID)
{
    ClassesTbl::iterator entry = m_classes.find(classID);
    if (entry == m_classes.end())
        return (IClassInfo*)NULL;
    else
        return entry->second.classInfo;
    
}

void Manager::addIIDMappings(const IIDMappingData& m)
{
    for (IIDMappingData::const_iterator i = m.begin(); i != m.end(); i++)
    {
        IIDsTbl::iterator e = m_iids.find(i->first);
        if (e != m_iids.end())
        {
            assert(e->second == i->second);
              // NB: such situation is not supposed to ever happen because
              //     the IID is made from name using MD5 hash. If it happens,
              //     we are in *DEEP* trouble, hence assert and not
              //     "return false;"
        }
        else
        {
            m_iids[i->first] = i->second;
        }    
    }    
}

Ptr<IInterfaceInfo> Manager::getInterfaceInfo(const IID& iid)
{
    IIDsTbl::iterator i = m_iids.find(iid);
    if (i == m_iids.end())
        return (IInterfaceInfo*)NULL;
    else
        return (IObject*)(privateData->typeInfoReg.createTypeInfo(i->second));
}


Ptr<IObject> Manager::createInstance(CID classID)
{
    ClassesTbl::iterator entry = m_classes.find(classID);

    if (entry == m_classes.end() || entry->second.factory == NULL)
        return (IObject*)NULL;
    else
        return entry->second.factory->create();
}

CIDsList Manager::enumInterfaceImplementations(const IID& iface)
{
   CIDsList list;
   
   for (ClassesTbl::iterator i = m_classes.begin(); i != m_classes.end(); i++)
   {
       if (i->second.classInfo->implementsInterface(iface))
           list.push_back(i->second.classInfo->getCID());
   }
   
   return list;
}


void Manager::loadClasses(ILoader *loader, const string& path)
{
    std::vector<std::string> files;
    std::vector<std::string>::const_iterator f;
    enumDirectory(path, loader->getExtensions(), files);
        
    for (f = files.begin(); f != files.end(); f++)
    {
        if (loader->canLoad(*f))
            loader->load(*f);
    }
}

void Manager::addModulesPath(const string& path)
{
    if (find(m_modulesPaths.begin(),
             m_modulesPaths.end(), path) != m_modulesPaths.end())
        return;

    // NB: This is the reason why we must use std::list and not std::vector for
    //     m_loaders -- new elements may be added into the list in this loop
    //     and such operation may invalidate std::vector's iterators
    for (LoadersList::iterator l = m_loaders.begin(); l != m_loaders.end(); l++)
    {
        loadClasses(*l, path);
    }
    // NB: add it to the list as last thing, so that one directory isn't loaded
    //     twice with same loader
    m_modulesPaths.push_back(path);
}

static inline string getExtension(const string& filename)
{
    size_t pos = filename.rfind('.');
    if (pos == string::npos)
        return string();
    else
        return filename.substr(pos);
}

bool Manager::loadModule(const string& filename)
{
    string fext(getExtension(filename));
    for (LoadersList::iterator l = m_loaders.begin(); l != m_loaders.end(); l++)
    {
        ILoader::ExtensionsList ext((*l)->getExtensions());
        if (ext.size() > 0)
        {
            bool ct = false;
            for (ILoader::ExtensionsList::const_iterator e = ext.begin();
                    e != ext.end(); e++)
            {
                if (fext == *e)
                {
                    ct = true;
                    break;
                }
            }
            if (!ct) continue;
        }
        if ((*l)->canLoad(filename))
        {
            if ((*l)->load(filename))
                return true;
        }
    }
    return false;
}

void Manager::addLoader(ILoader *loader)
{
    // NB: This is the reason why we must use std::list and not std::vector for
    //     m_modulesPaths -- new elements may be added into the list in this
    //     loop and such operation may invalidate std::vector's iterators
    for (list<string>::iterator path = m_modulesPaths.begin();
         path != m_modulesPaths.end(); path++)
    {
        loadClasses(loader, *path);
    }
    // NB: add it to the list as last thing, so that one directory isn't loaded
    //     twice with same loader
    m_loaders.push_back(loader);
}



UPF_IMPLEMENT_CLASS(upf::impl::Manager)
{
    UPF_INTERFACE(upf::IManager)
    UPF_PROPERTY("PartOf",      "UPF core")
    UPF_PROPERTY("Author",      "Vaclav Slavik")
}


} }
