
#include "config.h"
#include "upf/upf.h"
#include "upf/ILoader.h"
#include "upf/ILog.h"

#include <stdio.h>
#include "python-helper.h"

#ifdef __UPF_UNIX__
    #include "../upf_dlfcn.h"
    #define PY_CALL(x) (m_pythonAPI->x)
#else
    #define PY_CALL(x) x
#endif

using namespace std;
using namespace upf;

static void logError(const string& str)
{
    Ptr<ILog> log = getServiceManager()->resolve(SERVICE_LOG);
    log->log("upf", ILog::LOG_ERROR, str);
}

namespace upf { namespace impl {


class PythonLoader : public ILoader
{
public:
    PythonLoader();
    virtual ~PythonLoader();
    ExtensionsList getExtensions();
    bool load(const string& filename);
    bool canLoad(const string& filename);
    string getDescription();
    void init();

private:
    bool m_initialized;
    bool m_ok;
    bool m_ownsPython;

#ifdef __UPF_UNIX__
    void           *m_helperDLL;
    upf_python_api *m_pythonAPI;
#endif

    UPF_DECLARE_CLASS(PythonLoader)
};

PythonLoader::PythonLoader() 
    : m_initialized(false), m_ok(true), m_ownsPython(false),
      m_helperDLL(NULL), m_pythonAPI(NULL)
{
}

PythonLoader::~PythonLoader()
{
    if (m_initialized && m_ownsPython && m_ok)
    {
        PY_CALL(Py_Finalize)();
        m_ownsPython = m_initialized = false;
    }

#ifdef __UPF_UNIX__
    if (m_helperDLL)
        dlclose(m_helperDLL);
#endif
}

void PythonLoader::init()
{
    m_initialized = true;

#ifdef __UPF_UNIX__
    // NB: It *must* use RTLD_GLOBAL, see the comment in python-helper.h
    //     for explanation:
    m_helperDLL = dlopen(HELPER_FILENAME, RTLD_GLOBAL | RTLD_NOW);
    if (!m_helperDLL)
    {
        m_ok = false;
        logError("Can't load Python helper: " HELPER_FILENAME);
    }
    else
    {
        __upf_getPythonAPI_Func func =
            (__upf_getPythonAPI_Func)dlsym(m_helperDLL, __upf_getPythonAPI_Name);
        if (func == NULL)
        {
            m_ok = false;
            logError("Can't load Python helper: " HELPER_FILENAME);
        }
        else
            m_pythonAPI = func();
    }
    if (!m_ok) return;
#endif

    if (!PY_CALL(Py_IsInitialized)())
    {
        m_ownsPython = true;
        PY_CALL(Py_Initialize)();
        
        static const char *pycode_init =
           "import upf._impl\n"
           "upf._impl.init_from_python_loader()\n";
        if (PY_CALL(PyRun_SimpleString)((char*)pycode_init) != 0)
        {
            logError("Can't initialize Python interpreter.");
            m_ok = false;
            PY_CALL(Py_Finalize)();
        }
    }
}

ILoader::ExtensionsList PythonLoader::getExtensions()
{
    ExtensionsList list;
    if (m_ok)
        list.push_back(".py");
    return list;
}

bool PythonLoader::canLoad(const string& filename)
{
    return true;
}

bool PythonLoader::load(const string& filename)
{
    if (!m_initialized)
        init();
    if (!m_ok) return false;

    char pycode_buf[4096];

    snprintf(pycode_buf, 4096,
             "import upf._impl\n"
             "upf._impl.load_plugin('%s')\n",
             filename.c_str());
    if (PY_CALL(PyRun_SimpleString)(pycode_buf) != 0)
    {
        logError(string("Error when executing Python script: ") + filename);
        return false;
    }
    return true;
}

string PythonLoader::getDescription()
{
   return "Python script";
}



} } // end of namespace


UPF_IMPLEMENT_CLASS(upf::impl::PythonLoader)
{
    UPF_INTERFACE(upf::ILoader)
    UPF_PROPERTY("PartOf",      "UPF core")
    UPF_PROPERTY("Author",      "Vaclav Slavik")
    UPF_PROPERTY("Description", "Python modules loader")
}

UPF_DLL_MODULE()
{
    UPF_EXPORTED_CLASS(upf::impl::PythonLoader)
}
