
#include "upf/upf.h"

using namespace std;

#include "private.h"
#include <algorithm>
#include "upf_endian.h"

#include <stdio.h>
#if defined(_MSC_VER) || defined(__MINGW32__)
#define snprintf _snprintf
#endif

// --------------------------------------------------------------------------
// Typeinfo stream reader:
// --------------------------------------------------------------------------

namespace upf { namespace impl {

/** This class is a simple interface to stream of bytes. It's only useful
    features compater to raw pointer are that it takes care of little/big
    endian issues and that it maintains pointer to current read position.
 */
class TypeInfoStreamReader
{
public:
    TypeInfoStreamReader(const upf_uint8_t *stream) : m_stream(stream) {}

    upf_uint32_t read_uint32()
    {
        upf_uint32_t v;
        memcpy(&v, m_stream, 4);
        m_stream += 4;
        return _UPF_SWAP_NET32(v);
    }

    upf_uint16_t read_uint16()
    {
        upf_uint16_t v;
        memcpy(&v, m_stream, 2);
        m_stream += 2;
        return _UPF_SWAP_NET16(v);
    }

    upf_uint8_t read_uint8()
    {
        return *(m_stream++);
    }
    
    void read_data(void *mem, size_t len)
    {
        memcpy(mem, m_stream, len);
        m_stream += len;
    }
    
    void read_string(string& str)
    {
        upf_uint16_t lng = read_uint16();
        if (lng < 512)
        {
            char buf[512];
            read_data(buf, lng);
            buf[lng] = 0;
            str = buf;
        }
        else
        {
            char *buf = new char[lng+1];
            read_data(buf, lng);
            buf[lng] = 0;
            str = buf;
            delete[] buf;
        }
    }
    
    void read_iid(IID& iid)
    {
        iid.data1 = read_uint32();
        iid.data2 = read_uint32();
        iid.data3 = read_uint32();
        iid.data4 = read_uint32();
        iid.version_major = read_uint16();
        iid.version_minor = read_uint16();
    }
    
    void skip(size_t len)
    {
        m_stream += len;
    }
    
    const upf_uint8_t *get_state() { return m_stream; }
    void set_state(const upf_uint8_t *s) { m_stream = s; }

private:
    const upf_uint8_t *m_stream;
};

} } // namespace

// --------------------------------------------------------------------------
// ITypeInfo (and derived interfaces) implementation classes:
// --------------------------------------------------------------------------

namespace {

template<class T> struct TypeInfoTraits;

template<> struct TypeInfoTraits<upf::impl::InterfaceInfo>
{
    static const char* impl_name() { return "upf.impl.InterfaceInfo"; }
    typedef upf::IInterfaceInfo iface_type;
};

template<> struct TypeInfoTraits<upf::impl::TypedefInfo>
{
    static const char* impl_name() { return "upf.impl.TypedefInfo"; }
    typedef upf::ITypedefInfo iface_type;
};

template<> struct TypeInfoTraits<upf::impl::NativeTypeInfo>
{
    static const char* impl_name() { return "upf.impl.NativeTypeInfo"; }
    typedef upf::INativeTypeInfo iface_type;
};

}

namespace upf { namespace impl {

template<class T> void _TypeInfo<T>::init(TypeInfoStreamReader& stream)
{
    m_kind = (ITypeInfo::Kind)stream.read_uint8();
    stream.read_string(m_name);
}

UPF_IMPLEMENT_CLASS(upf::impl::TypeInfo)
{
    UPF_INTERFACE(upf::ITypeInfo)
    UPF_PROPERTY("PartOf",      "UPF core")
    UPF_PROPERTY("Author",      "Vaclav Slavik")
}

void InterfaceInfo::init(TypeInfoStreamReader& stream)
{
    _TypeInfo<IInterfaceInfo>::init(stream);
    stream.read_iid(m_iid);

    string name;
    upf_uint16_t i, p, cnt, cntp;

    // super classes:
    cnt = stream.read_uint16();
    m_supers.reserve(cnt);
    for (i = 0; i < cnt; i++)
    {
        stream.read_string(name);
        Ptr<IInterfaceInfo> ifc(privateData->typeInfoReg.createTypeInfo(name));
        m_supers.push_back(ifc);
    }

    // methods:
    cnt = stream.read_uint16();
    m_methods.reserve(cnt);
    for (i = 0; i < cnt; i++)
    {
        Method method;
        stream.read_string(method.name);
        stream.read_string(name);
        method.return_type = privateData->typeInfoReg.createTypeInfo(name);

        cntp = stream.read_uint16();
        method.params.reserve(cntp);
        for (p = 0; p < cntp; p++)
        {
            Parameter param;
            stream.read_string(name);
            param.type = privateData->typeInfoReg.createTypeInfo(name);
            stream.read_string(param.name);
            param.is_in = stream.read_uint8();
            param.is_out = stream.read_uint8();
            method.params.push_back(param);            
        }
        m_methods.push_back(method);
    }

    // flags:
    m_singleImpl = stream.read_uint8();
}

UPF_IMPLEMENT_CLASS(upf::impl::InterfaceInfo)
{
    UPF_INTERFACE(upf::IInterfaceInfo)
    UPF_PROPERTY("PartOf",      "UPF core")
    UPF_PROPERTY("Author",      "Vaclav Slavik")
}

void TypedefInfo::init(TypeInfoStreamReader& stream)
{
    _TypeInfo<ITypedefInfo>::init(stream);
    switch (stream.read_uint8())
    {
        case 1:
            m_seq = true;
            break;
        case 2:
            m_seq = false;
            break;
        default:
            assert(0 /*corrupted input data!*/);
    }
    stream.read_string(m_type);
}

UPF_IMPLEMENT_CLASS(upf::impl::TypedefInfo)
{
    UPF_INTERFACE(upf::ITypedefInfo)
    UPF_PROPERTY("PartOf",      "UPF core")
    UPF_PROPERTY("Author",      "Vaclav Slavik")
}

UPF_IMPLEMENT_CLASS(upf::impl::NativeTypeInfo)
{
    UPF_INTERFACE(upf::INativeTypeInfo)
    UPF_PROPERTY("PartOf",      "UPF core")
    UPF_PROPERTY("Author",      "Vaclav Slavik")
}


template<class T> class TypeInfoFactory
{
    typedef typename TypeInfoTraits<T>::iface_type iface_type;
public:
    static Ptr<ITypeInfo> create(TypeInfoStreamReader& stream)
    {
        Ptr<iface_type> obj = upf::create(TypeInfoTraits<T>::impl_name());
        T *impl = static_cast<T*>((iface_type*)obj);
        impl->init(stream);
        return (ITypeInfo*)obj;
    }
};

// --------------------------------------------------------------------------
// TypeInfoRegistry
// --------------------------------------------------------------------------


// NB: We choose lazy construction approach here. The ITypeInfo object for
//     given type is not created at registration time but rather when it
//     is requested by the app. The rationale is that the information is 
//     already in the memory (as read-only data of typeinf stream embedded
//     in loaded modules) and there's no point in creating another 
//     representation of same data, especially when it is most probably not
//     going to be used anyway (TypeInfo stuff is only useful for runtime
//     reflection support and for interfacing interpreted languages such
//     as Python or Java).

Ptr<ITypeInfo> TypeInfoRegistry::createTypeInfo(const string& name)
{
    Registry::iterator i = m_registry.find(name);
    if (i == m_registry.end())
        return (ITypeInfo*)NULL;

    if (i->second.constructed == NULL)
    {
        ITypeInfo::Kind kind = (ITypeInfo::Kind)(i->second.stream[0]);
        TypeInfoStreamReader stream(i->second.stream);

        switch (kind)
        {
            case ITypeInfo::TYPE_INTERFACE:
                i->second.constructed = 
                    TypeInfoFactory<InterfaceInfo>::create(stream);
                break;

            case ITypeInfo::TYPE_NATIVE_TYPE:
                i->second.constructed = 
                    TypeInfoFactory<NativeTypeInfo>::create(stream);
                break;

            case ITypeInfo::TYPE_TYPEDEF:
                i->second.constructed = 
                    TypeInfoFactory<TypedefInfo>::create(stream);
                break;

            default:
                //assert(0/*invalid type kind!!*/);
                // FIXME -- not everything implemented yet
                return (ITypeInfo*)NULL;
        }
    }
    
    return i->second.constructed;
}

bool TypeInfoRegistry::registerType(const string& id, const upf_uint8_t *chunk, size_t len)
{
    Registry::iterator i = m_registry.find(id);
    if (i == m_registry.end())
    {
        m_journal.push_back(JournalData(id, TypeData(chunk, len)));
        return true;
    }
    else
    {
        if (len != i->second.len)
            return false;
        if (memcmp(i->second.stream, chunk, len) != 0)
            return false;
        // FIXME - use logging interface to report exact cause of mismatch
        return true;
    }
}

void TypeInfoRegistry::rollback()
{
    m_journal.clear();
}

void TypeInfoRegistry::commit()
{
   for (Journal::iterator i = m_journal.begin(); i != m_journal.end(); i++)
   {
       m_registry[i->id] = i->data;
   }
    m_journal.clear();
}


// --------------------------------------------------------------------------
// Typeinfo stream interpreters/factories:
// --------------------------------------------------------------------------

int merge_typeinfo(const upf_uint8_t *typeinf)
{
    IIDMappingData iid_mappings;
    TypeInfoStreamReader ti(typeinf);
    upf_uint32_t len = ti.read_uint32();
    
    while (len)
    {
        const upf_uint8_t *state = ti.get_state();
        
        upf_uint8_t kind = ti.read_uint8();
        string id;
        ti.read_string(id);
        
        if ((ITypeInfo::Kind)kind == ITypeInfo::TYPE_INTERFACE)
        {
            upf_IID iid;
            ti.read_iid(iid);
            char buf[13]; // each %i is max 5 chars (16bit)
            snprintf(buf, 13, "/%i.%i", iid.version_major, iid.version_minor);
            id += buf;
            iid_mappings.push_back(IIDMappingData::value_type(iid, id));
        }

        if (!privateData->typeInfoReg.registerType(id, state, len))
        {
            privateData->typeInfoReg.rollback();
            return 0;
        }

        ti.set_state(state);
        ti.skip(len);
        
        len = ti.read_uint32();
    }

    privateData->manager->addIIDMappings(iid_mappings);
    privateData->typeInfoReg.commit();

    return 1;
}



} }
