
#ifndef __UPF_TYPEINFO_H__
#define __UPF_TYPEINFO_H__

#include "upf/upf.h"
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "upf/ITypeInfo.h"

#include <list>

namespace upf { namespace impl {

// Visual C++ needs this for some misterious reason :(
using std::string;
using std::pair;
using std::list;

// --------------------------------------------------------------------------
// ITypeInfo (and derived interfaces) implementation classes:
// --------------------------------------------------------------------------

class TypeInfoStreamReader;

template<class Parent> class _TypeInfo : public Parent
{
public:
    void init(TypeInfoStreamReader& stream);
    virtual ~_TypeInfo() {}

    ITypeInfo::Kind getKind() { return m_kind; }
    string getName() { return m_name; }

protected:
    ITypeInfo::Kind  m_kind;
    string           m_name;
};

class TypeInfo : public _TypeInfo<ITypeInfo>
{
public:
    void init(TypeInfoStreamReader& stream)
        { _TypeInfo<ITypeInfo>::init(stream); }
    UPF_DECLARE_CLASS(TypeInfo)
};


class InterfaceInfo : public _TypeInfo<IInterfaceInfo>
{
public:
    void init(TypeInfoStreamReader& stream);
    IID getIID() { return m_iid; }
    MethodList getMethods() { return m_methods; }
    InterfaceInfoList getSuperInterfaces() { return m_supers; }
    bool isSingleImpl() { return m_singleImpl; }

protected:
    IID m_iid;
    MethodList m_methods;
    InterfaceInfoList m_supers;
    bool m_singleImpl;

    UPF_DECLARE_CLASS(InterfaceInfo)
};


class TypedefInfo : public _TypeInfo<ITypedefInfo>
{
public:
    void init(TypeInfoStreamReader& stream);
    string getType() { return m_type; }
    bool isSequence() { return m_seq; }
    
protected:
    string m_type;
    bool m_seq;

    UPF_DECLARE_CLASS(TypedefInfo)
};

class NativeTypeInfo : public _TypeInfo<INativeTypeInfo>
{
    UPF_DECLARE_CLASS(NativeTypeInfo)
};



// --------------------------------------------------------------------------
// TypeInfoRegistry
// --------------------------------------------------------------------------

typedef list< pair<IID,string> > IIDMappingData;

class TypeInfoRegistry
{
public:
    TypeInfoRegistry() {}
    ~TypeInfoRegistry() {}

    /// Constructs ITypeInfo instance for given type name
    Ptr<ITypeInfo> createTypeInfo(const string& name);

    /** Registers type identified by given chunk of typeinf stream.
        This function is capable of detecting type mismatches. When a new
        component is being registered into the system, it reports description
        of all types it uses. If any piece of this description does not exactly
        match descriptions provided by other components, we have a mismatch. We
        can't allow such situation because it could lead to malfunction due to
        different interpretations of data.
        
        @return false if there was a type mismatch.
     */
    bool registerType(const string& id, const upf_uint8_t *chunk, size_t len);
    
    /// Undo all changes made to the registry
    void rollback();

    /** Commit all changes since previuos commit() or rollback() call and make
        them permanent. 
     */
    void commit();

private:
    struct TypeData
    {
        TypeData() {}
        TypeData(const upf_uint8_t *_stream, size_t _len) : stream(_stream), len(_len) {}
        
        const upf_uint8_t *stream;
        size_t             len;
        Ptr<ITypeInfo>     constructed;
    };
    
    struct JournalData
    {
        JournalData() {}
        JournalData(const string& _id, const TypeData& _data) : id(_id), data(_data) {}
        string   id;
        TypeData data;
    };

    typedef StringHash<TypeData> Registry;
    typedef list<JournalData> Journal;
    
    Registry m_registry;
    Journal m_journal; // this will be copied to m_registry in commit()
};


} } // end of namespace

#endif
