
#include <upf/upf.h>

#include <typeinfo>
#include <time.h>
#include <string>
#include <iostream>

#include "stress_test.h"

using namespace std;
using namespace upf;


class StressTestImpl : public IStressTest
{
public:
    void simpleCall() {}
    void passString(const string& value) {}
    StringList returnList()
    {
        StringList list;
        for (int i = 0; i < 50; i++)
           list.push_back("some returned string...");
        return list;
    }
    void fillList(StringList& list)
    {
        list.clear();
        for (int i = 0; i < 50; i++)
           list.push_back("some returned string...");
    }
    void passObj(upf::IManager *obj) {}
    
    UPF_DECLARE_CLASS(StressTestImpl)
};

UPF_IMPLEMENT_CLASS(StressTestImpl)
{
    UPF_INTERFACE(IStressTest)
}


class Timer
{
public:
    void start() { m_clock = clock(); }
    void stop() { m_time = ((double)(clock()-m_clock))/CLOCKS_PER_SEC; }
    double getTime() const { return m_time; }

private:
    clock_t m_clock;
    double m_time;
};

void report(const char *title, const Timer& t1, const Timer& t2, double times)
{
    char buf[2048];
    sprintf(buf, "%1.0f", times);
    cout << "*** " << title << " (" << buf << " iterations):" << endl;
    sprintf(buf, "%20s %20s", "(native)", "(marchalled)");
    cout << "                   " << buf << endl;
    sprintf(buf, "%19.3fs %19.3fs    (%1.1fx slower)", 
            t1.getTime(), t2.getTime(), t2.getTime()/t1.getTime());
    cout << "    elapsed time:  " << buf << endl;
    sprintf(buf, "%20.0f %20.0f", times/t1.getTime(), times/t2.getTime());
    cout << "    times per sec: " << buf << endl << endl;
}

#define RUN_TEST(description, cmd, multiplier) \
    { \
        Timer t1, t2; \
        int i; \
        t1.start(); \
        for (i = 0; i < ITER_CNT * multiplier; i++) \
        { \
            obj1->cmd; \
        } \
        t1.stop(); \
        t2.start(); \
        for (i = 0; i < ITER_CNT * multiplier; i++) \
        { \
            obj2->cmd; \
        } \
        t2.stop(); \
        report(description, t1, t2, multiplier * ITER_CNT); \
    }

void runBenchmark()
{
    const int ITER_CNT = 20000;
    
    Ptr<IStressTest> obj1 = upf::create("StressTestImpl");
    Ptr<IStressTest> obj2;
    Ptr<IObject> obj3 = upf::create("StressTestImpl");
    
    // NB: this hack makes sure obj2 use marchaller calls instead of native.
    //     obj1 always uses native calls because it is in same object module
    //     as the app:
    obj3->_upf_getABIObject()->nativeContext = upf_middleware->funcs.get_context(NULL);
    obj2 = obj3; // so that proxy is created
    
    cout << "native access type is " << typeid(*obj1).name() << endl;
    cout << "marchalled access type is " << typeid(*obj2).name() << endl;
    cout << endl;

    string str("dummy string");
    string longstr("vvery long string very long string very long string very long string very long string ery long string very long string very long string very long string very long string very long string very long string very long string very long string very long string very long string very long string ery long string very long string very long string very long string very long string very long string very long string ");
    StringList strlist;
    Ptr<IManager> obj = upf::getManager();

    RUN_TEST("simple call", simpleCall(), 100)
    RUN_TEST("one string argument", passString(str), 10)
    RUN_TEST("one long string argument", passString(longstr), 10)
    RUN_TEST("pass interface ptr", passObj(obj), 10)
    RUN_TEST("out list argument", fillList(strlist), 0.1)
    RUN_TEST("return string list", returnList(), 0.1)
}

int main()
{
    upf::init();
    UPF_REGISTER_CLASS(StressTestImpl);
    runBenchmark();
    upf::done();
}
